# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Command-line interface to rapport.
"""

from __future__ import print_function

import argparse
import datetime
import getpass
import os
import sys

import rapport
import rapport.config
import rapport.email
import rapport.plugin
import rapport.report
import rapport.timeframe


rapport.config.load()
rapport.plugin.discover()


class RapportHelpFormatter(argparse.HelpFormatter):
    def start_section(self, heading):
        # Title-case the headings
        heading = "{0}{1}".format(heading[0].upper(), heading[1:])
        super(RapportHelpFormatter, self).start_section(heading)


class CLI(object):
    def __init__(self):
        self.plugins = rapport.plugin.init_from_config()

    def create(self, args):
        if args.start:
            timeframe = rapport.timeframe.init("generic", start=args.start, end=args.end)
        elif args.current_month:
            timeframe = rapport.timeframe.init("current_month")
        elif args.current_week:
            timeframe = rapport.timeframe.init("current_week")
        elif args.week:
            timeframe = rapport.timeframe.init("week", week=args.week)
        elif args.month:
            timeframe = rapport.timeframe.init("month", month=args.month)
        elif args.recent_days:
            timeframe = rapport.timeframe.init("recent_days", days=args.recent_days)
        else:
            timeframe = rapport.timeframe.init_from_config()

        print("Timeframe: {0}".format(timeframe))
        report = rapport.report.create_report(self.plugins, timeframe)
        print("Work report {0}:\n{1}".format(report["date"], report["body"]))

    def list(self, args):
        for report in rapport.report.list_reports():
            print(report)

    def show(self, args):
        report = rapport.report.get_report(args.report)
        if args.raw:
            print(report)
        else:
            print(report["email.body.text"])

    def delete(self, args):
        for report in args.report:
            rapport.report.delete_report(report)

    def edit(self, args):
        rapport.report.edit_report(args.report, args.type)

   #def email_compose(self, args):
   #    pass

    def email_xdg(self, args):
        report = rapport.report.get_report(args.report)
        rapport.email.xdg_compose(to=rapport.config.get("email", "to"),
                                  subject=report["email.subject.text"],
                                  body=report["email.body.text"],
                                  cc=rapport.config.get("email", "cc"),
                                  bcc=rapport.config.get("email", "bcc"))

    def main(self):
        parser = argparse.ArgumentParser(prog="rapport",
                                         description="Work report generator for the lazy",
                                         epilog="See \"rapport help COMMAND\" for help on a specific command.",
                                         add_help=False,
                                         formatter_class=RapportHelpFormatter)
        parser.add_argument("--version", action="version", version=rapport.__version__,
                            help="Shows program's version and exits")
        parser.add_argument("-v", "--verbosity", action="count", help="Verbosity level (-v or -vv)")
        subparsers = parser.add_subparsers(title="Positional arguments", metavar="<subcommand>")

        parser_list = subparsers.add_parser("list", help="List existing work reports", formatter_class=RapportHelpFormatter)
        parser_list.set_defaults(func=self.list)

        parser_create = subparsers.add_parser("create", help="Create a new work report", formatter_class=RapportHelpFormatter)
        group_timeframe = parser_create.add_argument_group("Optional timeframe", description="Explicitly pick a timeframe in case you don't want to use the default.")
        group_timeframe.add_argument("-cm", "--current-month", action="store_true", help="Current month (until now)")
        group_timeframe.add_argument("-cw", "--current-week", action="store_true", help="Current week (until now)")
        group_timeframe.add_argument("-w", "--week", type=int, help="Week of year [1..52]")
        group_timeframe.add_argument("-m", "--month", type=int, help="Month of year [1..12]")
        group_timeframe.add_argument("-rd", "--recent-days", nargs="?", const=14, type=int, help="Recent days [1..] (default: 14)")
        group_generic_timeframe = parser_create.add_argument_group("Optional generic timeframe", description="Use in case the standard timeframes don't suit your needs. You can utilize 'date --iso-8601=seconds --utc' to get a date in UTC conforming to ISO 8601 if you have coreutils installed. Other means will work too, of course.")
        group_generic_timeframe.add_argument("-s", "--start", help="Start date [UTC ISO 8601]")
        group_generic_timeframe.add_argument("-e", "--end", nargs="?", default=datetime.datetime.utcnow(), help="End date [UTC ISO 8601] (default: now)")
        parser_create.set_defaults(func=self.create)

        parser_show = subparsers.add_parser("show", help="Display a specific work report", formatter_class=RapportHelpFormatter)
        parser_show.add_argument("-r", "--raw", action="store_true", help="Display the raw report data")
        parser_show.add_argument("report", nargs="?", default=None)
        parser_show.set_defaults(func=self.show)

        parser_edit = subparsers.add_parser("edit", help="Edit parts of a report prior to sending", formatter_class=RapportHelpFormatter)
        parser_edit.add_argument("report", nargs="?", default=None)
        parser_edit.add_argument("-t", "--type", default="email", choices=("email", "html"))
        parser_edit.add_argument("-ep", "--email-part", default="body", choices=("body", "subject"))
        parser_edit.set_defaults(func=self.edit)

        parser_delete = subparsers.add_parser("delete", help="Delete a work report", formatter_class=RapportHelpFormatter)
        parser_delete.add_argument("report", nargs="+", help="The report to delete")
        parser_delete.set_defaults(func=self.delete)

       #parser_email_compose = subparsers.add_parser("email-compose", formatter_class=RapportHelpFormatter)
       #parser_email_compose.set_defaults(func=self.email_compose)
        parser_email_xdg = subparsers.add_parser("email-compose-xdg", help="Use xdg-email to compose, i.e. your preferred mailer under KDE/GNOME/XFCE/etc.", formatter_class=RapportHelpFormatter)
        parser_email_xdg.add_argument("report", nargs="?", default=None)
        parser_email_xdg.set_defaults(func=self.email_xdg)

        parser_help = subparsers.add_parser("help", help="Show this help", formatter_class=RapportHelpFormatter)
        parser_help.set_defaults(func=lambda args: parser.print_help())
       #subparser.add_argument("-h", "--help", action="help",
       #                       help=argparse.SUPPRESS)

        args = parser.parse_args()
        args.func(args)

        #TODO: This is beginning to look ugly. Maybe use oslo.config or sth.:
        if args.verbosity > rapport.config.get_int("rapport", "verbosity"):
            rapport.config.set("rapport", "verbosity", args.verbosity)


def main():
   #try:
        CLI().main()
   #except Exception as e:
   #    print(e, file=sys.stderr)
   #    sys.exit(1)


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
E-Mail functionality.
"""

import subprocess

import rapport.config


def compose():
    raise NotImplementedError()


def send():
    raise NotImplementedError()


def xdg_compose(to, subject, body=None, cc=None, bcc=None):
    """Use xdg-email to compose in an X environment.

    Needs xdg-utils and a running X session. Works with GNOME, KDE,
    MATE, XFCE, ...
    """
    command = ["xdg-email", "--utf8", "--subject", subject]

    if body:
        command += ["--body", body]
    if cc:
        if type(cc) is list:
            cc = ", ".join(cc)
        command += ["--cc", cc]
    if bcc:
        if type(bcc) is list:
            bcc = ", ".join(bcc)
        command += ["--bcc", bcc]
    if type(to) is list:
        to = ", ".join(to)
    command.append(to)
    return subprocess.call(command)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import datetime
import glob
import os
import shutil
import subprocess
import sys
if sys.version_info > (3, 3):
    import concurrent.futures as futures
else:
    import futures
import traceback

import jinja2

import rapport.config
import rapport.template
import rapport.util


def _get_reports_path(report=None):
    path_parts = ["~", ".rapport", "reports"]
    if report:
        path_parts.append(report)
    return os.path.expanduser(os.path.join(*path_parts))


def list_reports():
    """Returns a list of created reports.
    """
    return sorted(os.listdir(_get_reports_path()))


def get_report(report=None):
    """Returns details of a specific report
    """
    if not report:
        report = list_reports()[-1:][0]
    report_path = _get_reports_path(report)
    report_dict = {"report": report}
    for filename in os.listdir(report_path):
        with open(os.path.join(report_path, filename), "r") as f:
            report_dict[filename] = f.read()
    return report_dict


def edit_report(report=None, type="email", email_part="body"):
    if not report:
        report = list_reports()[-1:][0]
    report_path = _get_reports_path(report)
    editor = os.getenv("EDITOR", "vi")
    if type == "email":
        report_file = "{0}.{1}.text".format(type, email_part)
    elif type == "html":
        report_file = "index.html"
    subprocess.call([editor, os.path.join(report_path, report_file)])


def create_report(plugins, timeframe):
    report_date_string = timeframe.end.strftime(rapport.util.ISO8610_FORMAT)
    report_path = _get_reports_path(report_date_string)
    if not os.path.exists(report_path):
        os.makedirs(report_path)

    # Execute all plugins in parallel and join on results:
    results = {}
    with futures.ThreadPoolExecutor(max_workers=4) as executor:
        plugin_futures = dict((executor.submit(p.collect, timeframe), p) for p in plugins)
        for future in futures.as_completed(plugin_futures):
            plugin = plugin_futures[future]
            try:
                if rapport.config.get_int("rapport", "verbosity") >= 2:
                    print("Result for {0}: {1}".format(plugin.alias, future.result()))
                tmpl = rapport.template.get_template(plugin, "text")
                if tmpl:
                    results[plugin] = tmpl.render(future.result())
            except jinja2.TemplateSyntaxError as e:
                print >>sys.stderr, "Syntax error in plugin {0} at {1} line {2}: {3}".format(plugin, e.name, e.lineno, e.message)                
            except Exception as e:
                print("Failed plugin {0}:{1}: {2}!".format(plugin, plugin.alias, e), file=sys.stderr)
                print(traceback.format_exc(), file=sys.stderr)

    results_dict = {"login": rapport.config.get("user", "login"),
                    "date": report_date_string,
                    "plugins": plugins,
                    "results": results}

    # Render mail body template:
    template_email_body = rapport.template.get_template("body", type="email")
    email_body = template_email_body.render(results_dict)
    email_body_file = os.path.join(report_path, "email.body.text")
    with open(email_body_file, "w") as report:
        report.write(email_body)

    # We can re-use the e-mail body as the general report body:
    results_dict["body"] = email_body

    # Render mail subject template:
    template_email_subject = rapport.template.get_template("subject", type="email")
    email_subject = template_email_subject.render(results_dict)
    email_subject_file = os.path.join(report_path, "email.subject.text")
    with open(email_subject_file, "w") as report:
        report.write(email_subject)

    #TODO: Maybe even create a Result class and return that instead of a dict?
    return results_dict


def delete_report(report):
    """Delete report(s), supports globbing.
    """
    for path in glob.glob(os.path.join(_get_reports_path(), report)):
        shutil.rmtree(path)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import glob
import os
import re
import shutil
import subprocess
import sys
from distutils.core import Command


class PEP8Command(Command):
    description = "Run pep8 with custom options"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call(["pep8", "--repeat", "--show-source",
                        "--ignore=E501",
                        "--exclude=.coverage/.venv,.tox,build,dist,doc,*egg*", "."])


class PEP257Command(Command):
    description = "Run pep257 with custom options"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        subprocess.call("find rapport -type f -name \"*.py\" | xargs pep257", shell=True)


class CleanupCommand(Command):
    patterns = [".coverage", ".tox", ".venv", "build", "dist", "*.egg", "*.egg-info"]
    description = "Clean up project directory"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        for pattern in CleanupCommand.patterns:
            for f in glob.glob(pattern):
                if os.path.isdir(f):
                    shutil.rmtree(f, ignore_errors=True)
                else:
                    os.remove(f)


def get_cmdclass():
    """Dictionary of all distutils commands defined in this module.
    """
    return {"cleanup": CleanupCommand,
            "pep257": PEP257Command,
            "pep8": PEP8Command}


def parse_requirements(requirements_file='requirements.txt'):
    requirements = []
    with open(requirements_file, 'r') as f:
        for line in f:
            # For the requirements list, we need to inject only the portion
            # after egg= so that distutils knows the package it's looking for
            # such as:
            # -e git://github.com/openstack/nova/master#egg=nova
            if re.match(r'\s*-e\s+', line):
                requirements.append(re.sub(r'\s*-e\s+.*#egg=(.*)$', r'\1',
                                    line))
            # such as:
            # http://github.com/openstack/nova/zipball/master#egg=nova
            elif re.match(r'\s*https?:', line):
                requirements.append(re.sub(r'\s*https?:.*#egg=(.*)$', r'\1',
                                    line))
            # -f lines are for index locations, and don't get used here
            elif re.match(r'\s*-f\s+', line):
                pass
            # -r lines are for including other files, and don't get used here
            elif re.match(r'\s*-r\s+', line):
                pass
            # argparse is part of the standard library starting with 2.7
            # adding it to the requirements list screws distro installs
            elif line == 'argparse' and sys.version_info >= (2, 7):
                pass
            else:
                requirements.append(line.strip())
    return requirements



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Trello plugin.
"""

import trolly.client
import trolly.member

import rapport.plugin
import rapport.util


# This is the public rapport Trello api key, which is used by default if you don't
# supply your own. In general you wouldn't need to either.
_RAPPORT_API_KEY = "e8b72b9823082ad89dd7dfb40e8373bd"


class TrelloPlugin(rapport.plugin.Plugin):
    def __init__(self, api_key=_RAPPORT_API_KEY, user_auth_token=None, *args, **kwargs):
        super(TrelloPlugin, self).__init__(*args, **kwargs)

        #TODO: Ask the user for a token if he didn't provide one with trolly.authorise
        self._c = trolly.client.Client(api_key=_RAPPORT_API_KEY, user_auth_token=user_auth_token)

    def collect(self, timeframe):
        # Even though Trolly provides Board and Card classes, fetching
        # everything in one go should be faster. Also, we want to filter by
        # 'dateLastActivity', which would take too much round-trips with
        # Trolly's Card class.
        card_fields = "name,desc,closed,dateLastActivity,due"
        json = self._c.fetchJson("/members/{0}".format(self.login),
                                 query_params={"boards": "all",
                                               "cards": "all",
                                               "organizations": "all",
                                               "card_fields": card_fields})

        open_cards, closed_cards, due_cards = [], [], []
        for card in json["cards"]:
            last_activity = rapport.util.datetime_from_iso8601(card["dateLastActivity"])
            if timeframe.contains(last_activity):
                if card["closed"]:
                    closed_cards.append(card)
                else:
                    open_cards.append(card)
                if card["due"]:
                    due_cards.append(card)

        open_boards, closed_boards = [], []
        for board in json["boards"]:
            if board["closed"]:
                closed_boards.append(board)
            else:
                open_boards.append(board)

        return self._results({"boards": json["boards"],
                              "open_boards": open_boards,
                              "closed_boards": closed_boards,
                              "cards": json["cards"],
                              "open_cards": open_cards,
                              "closed_cards": closed_cards,
                              "due_cards": due_cards,
                              "organizations": json["organizations"]})


rapport.plugin.register("trello", TrelloPlugin)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import datetime
import sys

import rapport.config


class Timeframe(object):
    """Represents a period of time between a start and end time.

    :start: Start of timeframe (datetime object)
    :end: End of timeframe (datetime object)
    """
    def __init__(self, start, end):
        self._name = "Generic"
        self._start = start
        self._end = end

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end

    def contains(self, date):
        """Checks if a date is within a timeframe.

        :date: The date to check
        """
        return self._start <= date and date < self._end

    def __str__(self):
        """Returns a string representation of a timeframe.
        """
        return "{0} [{1} - {2}]".format(self._name, self._start.isoformat(), self._end.isoformat())


class CurrentWeekTimeframe(Timeframe):
    """Current week timeframe (in UTC).
    """
    def __init__(self):
        self._name = "Current week"
        self._end = datetime.datetime.utcnow()
        # Compute the day but reset the hours/minutes/seconds to zero,
        # we want the exact week's start:
        week_start = self._end - datetime.timedelta(days=self._end.weekday())
        self._start = datetime.datetime(year=week_start.year,
                                        month=week_start.month,
                                        day=week_start.day)


class WeekTimeframe(Timeframe):
    """N-th week of year timeframe (in UTC).

    :week: Week number (starting from 1)
    """
    def __init__(self, week=1):
        self._name = "Week"
        raise NotImplementedError()


class CurrentMonthTimeframe(Timeframe):
    """Current month timeframe (in UTC).
    """
    def __init__(self):
        self._name = "Current month"
        self._end = datetime.datetime.utcnow()
        self._start = datetime.datetime(year=self._end.year,
                                        month=self._end.month, day=1)


class MonthTimeframe(Timeframe):
    """N-th month of year timeframe (in UTC).

    :month: Month number (starting from 1)
    """
    def __init__(self, month=1):
        self._name = "Month"
        raise NotImplementedError()


class RecentDaysTimeframe(Timeframe):
    """Recent days timeframe (in UTC).
    """
    def __init__(self, days=14):
        self._name = "Recent days ({0})".format(days)
        self._end = datetime.datetime.utcnow()
        self._start = self._end - datetime.timedelta(days=days)


_TIMEFRAME_CATALOG = {"current_month": CurrentMonthTimeframe,
                      "current_week": CurrentWeekTimeframe,
                      "month": MonthTimeframe,
                      "week": WeekTimeframe,
                      "recent_days": RecentDaysTimeframe}


def init(name, *args, **kwargs):
    """Instantiate a timeframe from the catalog.
    """
    if name in _TIMEFRAME_CATALOG:
        if rapport.config.get_int("rapport", "verbosity") >= 2:
            print("Initialize timeframe {0}: {1} {2}".format(name, args, kwargs))
        try:
            return _TIMEFRAME_CATALOG[name](*args, **kwargs)
        except ValueError as e:
            print("Failed to initialize timeframe {0}: {1}!".format(name, e), file=sys.stderr)
    else:
        print("Failed to initialize timeframe {0}: Not in catalog!".format(name), file=sys.stderr)
        sys.exit(1)


def init_from_config():
    return init(rapport.config.get("timeframe", "default"))


def catalog():
    """Returns the list of registered timeframes.
    """
    return _TIMEFRAME_CATALOG.keys()



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Bugzilla plugin.
"""

# Avoid name clash with global "bugzilla" module from python-bugzilla:
from __future__ import absolute_import
import bugzilla

import rapport.plugin


class BugzillaPlugin(rapport.plugin.Plugin):
    def __init__(self, email, *args, **kwargs):
        super(BugzillaPlugin, self).__init__(*args, **kwargs)
        self.email = email

        url = "{0}/xmlrpc.cgi".format(self.url.geturl())

    def collect(self, timeframe):
        bz = bugzilla.Bugzilla(url=url)
        bz.login(user=self.login, password=self.password)

        open_bugs = bz.query({"assigned_to": self.email,
                              "bug_status": ["NEW", "ASSIGNED"]})
        closed_bugs = bz.query({"assigned_to": self.email,
                                "bug_status": ["CLOSED", "RESOLVED"],
                                "last_change_time": timeframe.start})

        return self._results({"email": self.email, "open_bugs": open_bugs,
                              "closed_bugs": closed_bugs})


rapport.plugin.register("bugzilla", BugzillaPlugin)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various utility functions.
"""

import datetime
import re
import site
import sys
import subprocess


_FIRST_CAP_RE = re.compile('(.)([A-Z][a-z]+)')
_ALL_CAP_RE = re.compile('([a-z0-9])([A-Z])')

ISO8610_FORMAT = "%Y-%m-%dT%H:%M:%S"
ISO8610_FORMAT_MICROSECONDS = "%Y-%m-%dT%H:%M:%S.%f"


def camelcase_to_underscores(word):
    """Converts a CamelCase word into an under_score word.

        >>> camelcase_to_underscores("CamelCaseCase")
        'camel_case_case'
        >>> camelcase_to_underscores("getHTTPResponseCode")
        'get_http_response_code'
    """
    s1 = _FIRST_CAP_RE.sub(r'\1_\2', word)
    return _ALL_CAP_RE.sub(r'\1_\2', s1).lower()


def silent_popen(args, **kwargs):
    """Wrapper for subprocess.Popen with suppressed output.

    STERR is redirected to STDOUT which is piped back to the
    calling process and returned as the result.
    """
    return subprocess.Popen(args,
                            stderr=subprocess.STDOUT,
                            stdout=subprocess.PIPE, **kwargs).communicate()[0]


def datetime_from_iso8601(date):
    """Small helper that parses ISO-8601 date dates.

        >>> datetime_from_iso8601("2013-04-10T12:52:39")
        datetime.datetime(2013, 4, 10, 12, 52, 39)
        >>> datetime_from_iso8601("2013-01-07T12:55:19.257")
        datetime.datetime(2013, 1, 7, 12, 55, 19, 257000)
    """
    format = ISO8610_FORMAT
    if date.endswith("Z"):
        date = date[:-1]  # Date date is UTC
    if re.match(".*\.\d+", date):
        # Date includes microseconds
        format = ISO8610_FORMAT_MICROSECONDS
    return datetime.datetime.strptime(date, format)


def under_virtualenv():
    return hasattr(sys, "real_prefix")


def getsitepackages():
    if hasattr(site, "getsitepackages"):
        return site.getsitepackages()
    else:
        # Workaround for older Python versions and some broken virtualenvs:
        if under_virtualenv():
            return []
        else:
            from distutils.sysconfig import get_python_lib
            return [get_python_lib(True)]



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Gerrit plugin.
"""

from __future__ import print_function

import json
from datetime import datetime

import paramiko

import rapport.plugin


class GerritPlugin(rapport.plugin.Plugin):
    def __init__(self, *args, **kwargs):
        super(GerritPlugin, self).__init__(*args, **kwargs)

        self._client = paramiko.SSHClient()
        self._client.load_system_host_keys()

    def _ssh_cmd(self, *args):
        """Execute a gerrit command over SSH.
        """
        command = "gerrit {0}".format(" ".join(args))
        _, stdout, stderr = self._client.exec_command(command)
        return (stdout.readlines(), stderr.readlines())

    def _ssh_query(self, *args):
        """Execute a gerrit query over SSH and returns JSON-formatted data.
        """
        return self._ssh_cmd("query", "--format=JSON", *args)

    def collect(self, timeframe):
        self._client.connect(self.url.hostname, self.url.port, self.login)
        stdout, stderr = self._ssh_query("owner:{0}".format(self.login))
        self._client.close()

        changes = []
        if not stderr:
            for line in stdout[:-1]:  # Last line contains only download stats
                change = json.loads(line)
                if "lastUpdated" in change:
                    last_updated = datetime.utcfromtimestamp(
                        change["lastUpdated"])
                    if timeframe.contains(last_updated):
                        changes.append(change)
                else:
                    print("Change {0} is missing lastUpdated".format(change))

        return self._results({"changes": changes})


rapport.plugin.register("gerrit", GerritPlugin)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import sys
import traceback
try:
    import urllib.parse as urlparse
except ImportError:
    import urlparse

import rapport.config
import rapport.util


class Plugin(object):
    def __init__(self, alias=None, url=None, login=None, password=None):
        self.alias = alias
        self.url = url
        self.login = login
        self.password = password

        if alias == "":
            self.alias == self.__str__()

        if url and type(url) is not urlparse.ParseResult:
            self.url = urlparse.urlparse(url)

    def _results(self, dict={}):
        """Helper to merge a dict with cross-plugin defaults.

        All plugin sub-classes share some config values, i.e.  alias, url,
        login and password. This help should be used in the :collect: method
        of any Plugin implementation.

           >>> import rapport.plugin
           >>> c = rapport.plugin.Plugin(alias="a", url="http://example.com", login="u")
           >>> c._results()
           {'url': ParseResult(scheme='http', netloc='example.com', path='', params='', query='', fragment=''), 'alias': 'a', 'login': 'u', 'plugin': 'plugin'}
           >>> c._results({'mykey': 'mykey'})
           {'url': ParseResult(scheme='http', netloc='example.com', path='', params='', query='', fragment=''), 'alias': 'a', 'login': 'u', 'mykey': 'mykey', 'plugin': 'plugin'}
        """
        results = {"plugin": str(self),
                   "alias": self.alias,
                   "url": self.url,
                   "login": self.login}
        results.update(dict)
        return results

    def __str__(self):
        """Returns the class name in underscores.

        Additionally, for sub-classes, the suffix '_plugin' is split off.

            >>> import rapport.plugin
            >>> c = rapport.plugin.Plugin()
            >>> str(c)
            'plugin'
        """
        return rapport.util.camelcase_to_underscores(self.__class__.__name__).rsplit("_plugin")[0]

    def collect(self, timeframe):
        raise NotImplementedError()


def _get_plugin_dirs():
    """Return a list of directories where plugins may be located.
    """
    plugin_dirs = [
        os.path.expanduser(os.path.join("~", ".rapport", "plugins")),
        os.path.join("rapport", "plugins")  # Local dev tree
    ]
    return plugin_dirs


def _path_to_module(path):
    """Translates paths to *.py? files into module paths.

        >>> _path_to_module("rapport/bar.py")
        'rapport.bar'
        >>> _path_to_module("/usr/lib/rapport/bar.py")
        'rapport.bar'
    """
    # Split of preceeding path elements:
    path = "rapport" + path.split("rapport")[1]
    # Split of ending and replace os.sep with dots:
    path = path.replace(os.sep, ".").rsplit(".", 1)[0]
    return path


def discover():
    """Find and load all available plugins.
    """
    plugin_files = []

    for plugin_dir in _get_plugin_dirs():
        if os.path.isdir(plugin_dir):
            for plugin_file in os.listdir(plugin_dir):
                if plugin_file.endswith(".py") and not plugin_file == "__init__.py":
                    plugin_files.append(os.path.join(plugin_dir, plugin_file))

    if rapport.config.get_int("rapport", "verbosity") >= 2:
        print("Found plugin modules: {0}".format(plugin_files))

    for plugin_file in plugin_files:
        if rapport.config.get_int("rapport", "verbosity") >= 2:
            print("Importing module {0}".format(_path_to_module(plugin_file)))
        __import__(_path_to_module(plugin_file))


_PLUGIN_CATALOG = {}


def register(name, klass):
    """Add a plugin to the plugin catalog.
    """
    if rapport.config.get_int("rapport", "verbosity") >= 1:
        print("Registered plugin: {0}".format(name))
    _PLUGIN_CATALOG[name] = klass


def init(name, *args, **kwargs):
    """Instantiate a plugin from the catalog.
    """
    if name in _PLUGIN_CATALOG:
        if rapport.config.get_int("rapport", "verbosity") >= 2:
            print("Initialize plugin {0}: {1} {2}".format(name, args, kwargs))
        try:
            return _PLUGIN_CATALOG[name](*args, **kwargs)
        except (ValueError, TypeError) as e:
            print("Failed to initialize plugin {0}: {1}!".format(name, e), file=sys.stderr)
    else:
        print("Failed to initialize plugin {0}: Not in catalog!".format(name), file=sys.stderr)


def init_from_config():
    plugins = []
    for plugin in rapport.config.plugins():
        plugins.append(init(**plugin))
    return list(filter(bool, plugins))


def catalog():
    """Returns the list of registered plugins.
    """
    return _PLUGIN_CATALOG.keys()



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Mediawiki plugin.
"""

import json

import requests

import rapport.plugin


class MediawikiPlugin(rapport.plugin.Plugin):
    def __init__(self, *args, **kwargs):
        super(MediawikiPlugin, self).__init__(*args, **kwargs)

    def _get(self, params={}):
        params.update({"format": "json", "action": "query"})
        response = requests.get(self.url.geturl(), params=params, auth=(self.login, self.password))
        return json.loads(response.text)["query"]

    def collect(self, timeframe):
        #TODO: Check back ugly "UTC" timestamp hack here:
        edits = self._get({"list": "usercontribs",
                                   "ucuser": "{0}".format(self.login),
                                   "ucstart": timeframe.end.isoformat() + "Z",
                                   "ucend": timeframe.start.isoformat() + "Z"})
        return self._results({"edits": edits["usercontribs"]})


rapport.plugin.register("mediawiki", MediawikiPlugin)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
:mod:`rapport` -- Work report generator for the lazy
====================================================

.. automodule:: rapport
   :synopsis: Work report generator for the lazy
.. moduleauthor:: Sascha Peilicke <saschpe@gmx.de>
"""

__author__ = "Sascha Peilicke <saschpe@gmx.de>"
__version_info__ = ("0", "1", "3")
__version__ = ".".join(__version_info__)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import shutil
try:
    import ConfigParser as configparser  # Py2
except ImportError:
    import configparser

import rapport.config


def _get_config_dirs():
    """Return a list of directories where config files may be located.

    The following directories are returned::

      ~/.rapport/
      /etc/rapport/
    """
    config_dirs = [
        os.path.expanduser(os.path.join("~", ".rapport")),
        os.path.join("/", "etc", "rapport"),
        os.path.abspath(os.path.join("rapport", "config"))
    ]
    return config_dirs


def find_config_files():
    """Return a list of default configuration files.
    """

    config_files = []

    for config_dir in _get_config_dirs():
        path = os.path.join(config_dir, "rapport.conf")
        if os.path.exists(path):
            config_files.append(path)

    return list(filter(bool, config_files))


def init_user():
    """Create and populate the ~/.rapport directory tree if it's not existing.

    Doesn't interfere with already existing directories or configuration files.
    """
    user_conf_dir = os.path.expanduser(os.path.join("~", ".rapport"))
    user_conf_file = os.path.join(user_conf_dir, "rapport.conf")

    if not os.path.exists(user_conf_dir):
        if rapport.config.get_int("rapport", "verbosity") >= 1:
            print("Create user directory {0}".format(user_conf_dir))
        os.makedirs(user_conf_dir)
    for subdir in ["plugins", "reports", "templates/plugin", "templates/email", "templates/web"]:
        user_conf_subdir = os.path.join(user_conf_dir, subdir)
        if not os.path.exists(user_conf_subdir):
            if rapport.config.get_int("rapport", "verbosity") >= 1:
                print("Create user directory {0}".format(user_conf_subdir))
            os.makedirs(user_conf_subdir)
        if subdir == "reports" and not (os.stat(user_conf_subdir).st_mode & 0o777) == 0o700:
            if rapport.config.get_int("rapport", "verbosity") >= 1:
                print("Set secure directory permissions for {0}".format(user_conf_subdir))
            os.chmod(user_conf_subdir, 0o700)
    if not os.path.exists(user_conf_file):
        if rapport.config.get_int("rapport", "verbosity") >= 1:
            print("Create user configuration {0}".format(user_conf_file))
        default_config = os.path.abspath(os.path.join(os.path.splitext(__file__)[0], "rapport.conf"))
        shutil.copyfile(default_config, user_conf_file)

    if not (os.stat(user_conf_file).st_mode & 0o777) == 0o600:
        if rapport.config.get_int("rapport", "verbosity") >= 1:
            print("Set secure file permissions for {0}".format(user_conf_file))
        os.chmod(user_conf_file, 0o600)


CONF = None


def load():
    global CONF
    config = configparser.SafeConfigParser()

    if not find_config_files():
        init_user()
    config.read(find_config_files()[0])
    CONF = config
    return CONF


def get(section, option, default=None):
    if CONF.has_option(section, option):
        return CONF.get(section, option)
    else:
        return default


def get_int(section, option, default=-1):
    return int(get(section, option, default))


def set(section, option, value):
    CONF.set(section, option, str(value))


def plugins():
    for section in CONF.sections():
        if section.startswith("plugin:"):
            name, alias = section.split(":")[1:]
            plugin = {"name": name, "alias": alias}
            for option in CONF.options(section):
                plugin[option] = CONF.get(section, option)
            yield plugin



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Twitter plugin.
"""

import tweepy

import rapport.plugin


class TwitterPlugin(rapport.plugin.Plugin):
    def __init__(self, consumer_key, consumer_secret, access_token, access_token_secret, *args, **kwargs):
        super(TwitterPlugin, self).__init__(*args, **kwargs)

        auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
        auth.set_access_token(access_token, access_token_secret)

        self.api = tweepy.API(auth)

    def _get(self, timeline, timeframe):
        #TODO: Implement proper pagination instead of fixed count:
        count = 50
        statuses = []
        for status in timeline(count=count):
            if timeframe.contains(status.created_at):
                statuses.append(status)
        return statuses

    def collect(self, timeframe):
        mentions = self._get(self.api.mentions_timeline, timeframe)
        tweets = self._get(self.api.user_timeline, timeframe)
        retweets = self._get(self.api.retweets_of_me, timeframe)

        return self._results({"mentions": mentions,
                              "tweets": tweets,
                              "retweets": retweets})


rapport.plugin.register("twitter", TwitterPlugin)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Open Build Service plugin.
"""

import lxml.etree
import requests

import rapport.plugin


class OpenBuildServicePlugin(rapport.plugin.Plugin):
    def __init__(self, *args, **kwargs):
        super(OpenBuildServicePlugin, self).__init__(*args, **kwargs)

    def _get_xml(self, url):
        response = requests.get(url, auth=(self.login, self.password))
        return lxml.etree.fromstring(response.text)

    def collect(self, timeframe):
        # Directly querying for all user's requests is killing the OBS api.
        # So we ask for all projects / packages where the user is involved in first.

        user_projects_url = "{0}/search/project/id?match=person/@userid='{1}'".format(self.url.geturl(), self.login)
        xml_root = self._get_xml(user_projects_url)
        user_projects = [project.get('name') for project in xml_root.iterfind(".//project")]

        #user_packages_url = "{0}/search/package/id?match=person/@userid='{1}'".format(self.url.geturl(), self.login)
        #xml_root = self._get_xml(user_projects_url)
        #user_packages = [package.values() for package in xml_root.iterfind(".//package")]

        #"https://api.opensuse.org/search/request/id?match=history/@who='saschpe'"

        for project in user_projects:
            requests = self._get_xml("{0}/request?view=collection&user={1}&project={2}".format(self.url.geturl(), self.login, project))


rapport.plugin.register("openbuildservice", OpenBuildServicePlugin)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Github plugin.
"""

import collections
import json
import re

import requests

import rapport.plugin
import rapport.util


class GithubPlugin(rapport.plugin.Plugin):
    def __init__(self, *args, **kwargs):
        super(GithubPlugin, self).__init__(*args, **kwargs)
        #TODO: Maybe rather use oauth?

    def _get_json(self, url):
        response = requests.get(url, auth=(self.login, self.password))
        link_url = None
        if "link" in response.headers:
            link_url = response.headers["link"]
            if link_url.startswith("<"):
                # If URL looks like this: '<https://api.github.com/user/$ID/events?page=2>; rel="next"'
                link_url = re.match('<([^>]+)>; rel="next"', link_url).groups()[0]
        return json.loads(response.text), link_url

    def collect(self, timeframe):
        url = "https://api.github.com/users/{0}/events".format(self.login)
        d = collections.defaultdict(list)

        # Paginate through the activity stream, mark first item in timeline
        # and last one to have a criteria for stopping pagination.
        while True:
            events_json, url = self._get_json(url)

            for event in events_json:
                created_at = rapport.util.datetime_from_iso8601(event["created_at"])
                if timeframe.contains(created_at):
                    d["events"].append(event)
                    if event["type"] == "CommitCommentEvent":
                        d["commit_comment_events"].append(event)
                    elif event["type"] == "CreateEvent":
                        d["create_events"].append(event)
                    elif event["type"] == "DeleteEvent":
                        d["delete_events"].append(event)
                    elif event["type"] == "ForkEvent":
                        d["fork_events"].append(event)
                    elif event["type"] == "GistEvent":
                        d["gist_events"].append(event)
                    elif event["type"] == "GollumEvent":
                        d["gollum_events"].append(event)
                    elif event["type"] == "IssuesEvent":
                        d["issues_events"].append(event)
                    elif event["type"] == "IssueCommentEvent":
                        d["issues_comment_events"].append(event)
                    elif event["type"] == "PullRequestEvent":
                        d["pull_request_events"].append(event)
                    elif event["type"] == "PullRequestReviewCommentEvent":
                        d["pull_request_review_comment_events"].append(event)
                    elif event["type"] == "PushEvent":
                        d["push_events"].append(event)
                    elif event["type"] == "TeamAddEvent":
                        d["team_add_events"].append(event)
                else:
                    if d["events"]:  # Found the last event inside the timeframe
                        url = None  # Don't fetch another page, got everything interesting
                        break  # No need to process the remaining events on this page

            if url is None:
                break  # Reached last page of activity stream, that's it :-)

        return self._results(d)


rapport.plugin.register("github", GithubPlugin)



# Copyright 2013 Sascha Peilicke
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os
import sys

import jinja2


def _get_template_dirs(type="plugin"):
    """Return a list of directories where templates may be located.
    """
    template_dirs = [
        os.path.expanduser(os.path.join("~", ".rapport", "templates", type)),
        os.path.join("rapport", "templates", type)  # Local dev tree
    ]
    return template_dirs


_JINJA2_ENV = {}


def init():
    for type in ["plugin", "email", "web"]:
        loader = jinja2.FileSystemLoader(_get_template_dirs(type))
        env = jinja2.Environment(loader=loader,
                                 extensions=["jinja2.ext.i18n", "jinja2.ext.loopcontrols"],
                                 line_statement_prefix="%%",
                                 line_comment_prefix="##",
                                 trim_blocks=True)
        env.install_null_translations(newstyle=False)
        _JINJA2_ENV[type] = env


def get_template(name, format="text", type="plugin"):
    if not _JINJA2_ENV:
        init()
    template_name = "{0}.{1}.jinja2".format(name, format)
    try:
        return _JINJA2_ENV[type].get_template(template_name)
    except jinja2.TemplateNotFound as e:
        print("Missing template {0}/{1}!".format(type, template_name), file=sys.stderr)
